/*
** File: grammar.y 
** Content: This file contains yacc (or bison) parser generator code
**          for processing the Well-Known Text format for geometry
**          described in OpenGIS Simle Features for SQL, Revision 1.0
** Author: Kurt Buehler, Open GIS Consortium, Inc.
** Date: 7/27/1998
**
** Revision History:
**     1.0 - Initial Release, 7/27/1998
**
*/

/*
** Copyright 1998, Open GIS Consortium, Inc.
**
** WHILE THE SOFTWARE IN THIS FILE IS BELIEVED TO BE ACCURATE, 
** THE OPEN GIS CONSORTIUM MAKES NO WARRANTY OF ANY KIND WITH REGARD TO 
** THIS MATERIAL INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
** MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE. The Open GIS
** Consortium shall not be liable for errors contained herein or for incidental
** or consequential damages in connection with the furnishing, performance or 
** use of this material.
**
** This document contains information, which is protected by copyright. All 
** Rights Reserved. Any part of this work covered by copyright herein may be 
** reproduced or used in any form or by any means graphic, electronic, or 
** mechanical, including photocopying, recording, taping, or information 
** storage and retrieval systems as long as this copyright notice remains.
**
** RESTRICTED RIGHTS LEGEND. Use, duplication, or disclosure by government is 
** subject to restrictions as set forth in subdivision (c)(1)(ii) of the Right 
** in Technical Data and Computer Software Clause at DFARS 252.227.7013
** 
** OpenGIS(R) is a trademark or registered trademark of Open GIS Consortium, 
** Inc. in the United States and in other countries.
*/

%{

#include "wkb.h"
#define YYERROR_VERBOSE 1
ParsedInput *global_parsed;

%}

%start parsed_input

%union {
    int ival;
    double real_val;
    char *string_val;
    Point *point_val; /* POINT */
    PointList *multipoint_val; /* MULTIPOINT */
    Linestring *linestring_val; /* LINESTRING */
    LinestringList *multilinestring_val; /* MULTILINESTRING */
    LinestringList *polygon_val; /* POLYGON */
    PolygonList *multipolygon_val; /* MULTIPOLYGON */
    GeometryCollection *geometry_collection_val; /* GEOMETRYCOLLECTION */
    Geometry *geometry_val; /* any geometry */
    ParsedInput *parsed_input_val; /* parsed input */
}

%token LPAREN
%token RPAREN
%token COMMA
%token REAL
%token POINT
%token LINESTRING
%token POLYGON
%token MULTIPOINT
%token MULTILINESTRING
%token MULTIPOLYGON
%token GEOMETRYCOLLECTION
%token GEOMETRY

%type <real_val> REAL
%type <point_val> pair
%type <point_val> point
%type <multipoint_val> vertex_list
%type <linestring_val> linestring
%type <linestring_val> exterior_ring
%type <multilinestring_val> interior_rings
%type <multilinestring_val> ring_list
%type <multilinestring_val> polygon
%type <multipoint_val> multipoint
%type <multipoint_val> pair_list
%type <multilinestring_val> linestring_list
%type <multilinestring_val> multilinestring
%type <multipolygon_val> polygon_list
%type <multipolygon_val> multipolygon
%type <geometry_collection_val> geometry_list
%type <geometry_collection_val> geometrycollection
%type <geometry_val> geometry
%type <parsed_input_val> parsed_input


%%
parsed_input: geometry
    {
		ParsedInput *parsed = (ParsedInput *)malloc(sizeof(ParsedInput));
	    parsed->type = GEOMETRY;
		parsed->value.geometry_val = $1;
		global_parsed = parsed;
		$$ = parsed;
	}
	        | geometrycollection
	{
		ParsedInput *parsed = (ParsedInput *)malloc(sizeof(ParsedInput));
	    parsed->type = GEOMETRYCOLLECTION;
		parsed->value.geometrycollection_val = $1;
		global_parsed = parsed;
		$$ = parsed;
	}

geometry : point 
    {
	    Geometry *g = (Geometry *)malloc(sizeof(Geometry));
		g->type = POINT;
		g->value.point_val = $1;
		$$ = g;
	}
         | linestring
    {
	    Geometry *g = (Geometry *)malloc(sizeof(Geometry));
		g->type = LINESTRING;
		g->value.linestring_val = $1;
		$$ = g;
	}
         | polygon
    {
	    Geometry *g = (Geometry *)malloc(sizeof(Geometry));
		g->type = POLYGON;
		g->value.polygon_val = $1;
		$$ = g;
	}
         | multipoint
    {
	    Geometry *g = (Geometry *)malloc(sizeof(Geometry));
		g->type = MULTIPOINT;
		g->value.multipoint_val = $1;
		$$ = g;
	}
         | multilinestring
    {
	    Geometry *g = (Geometry *)malloc(sizeof(Geometry));
		g->type = MULTILINESTRING;
		g->value.multilinestring_val = $1;
		$$ = g;
	}
         | multipolygon
    {
	    Geometry *g = (Geometry *)malloc(sizeof(Geometry));
		g->type = MULTIPOLYGON;
		g->value.multipolygon_val = $1;
		$$ = g;
	}

pair : REAL REAL
    {
		Point *p = (Point *)malloc(sizeof(Point));
		p->x = $1; p->y = $2;
		$$ = p;
	}

point : POINT LPAREN pair RPAREN
    { 
	    $$ = $3;
    }

vertex_list : pair
    {
		/* allocate a new point list item */
		PointList *p = (PointList *)malloc(sizeof(PointList));
		/* populate the point list item */
		p->point = $1;
		p->next = NULL;
		/* set vertex_list to the right thing */
		$$ = p;
	}
            | vertex_list COMMA pair
    {
		/* retrieve the pointer to the beginning of the list */
		PointList *old = $1;
		/* declare a pointer to do traversal */
		PointList *oldptr;
		/* allocate a new point list item */
	    PointList *p = (PointList *)malloc(sizeof(PointList));

		/* populate the point list item */
		p->point = $3;
		p->next = NULL;
		/* traverse the current list until we reach the end */
		oldptr = old;
		while (oldptr->next != NULL) oldptr = oldptr->next;
		/* link in the next point list item */
		oldptr->next = p;
		/* set vertex_list to the right thing */
        $$ = old;
	}

linestring : LINESTRING LPAREN vertex_list RPAREN
    {
	    $$ = $3;
    }

exterior_ring: LPAREN vertex_list RPAREN
    {
	    $$ = $2;
    }

interior_rings : LPAREN vertex_list RPAREN
    {
		/* allocate a new linestring list item */
		LinestringList *l = (LinestringList *)malloc(sizeof(LinestringList));
		/* populate the linestring list item */
		l->linestring = $2;
		l->next = NULL;
		/* set interior_rings to the right thing */
		$$ = l;
    }
               | interior_rings COMMA LPAREN vertex_list RPAREN
    {
		/* retrieve the pointer to the beginning of the list */
		LinestringList *old = $1;
		/* declare a pointer to do traversal */
		LinestringList *oldptr;
		/* allocate a new linestring list item */
		LinestringList *l = (LinestringList *)malloc(sizeof(LinestringList));

		/* populate the linestring list item */
		l->linestring = $4;
		l->next = NULL;
		/* traverse the current list until we reach the end */
		oldptr = old;
		while (oldptr->next != NULL) oldptr = oldptr->next;
		/* link in the next point list item */
		oldptr->next = l;
		/* set interior_rings to the right thing */
        $$ = old;
    }

ring_list : exterior_ring
	{
		/* allocate a new linestring list item */
		LinestringList *l = (LinestringList *)malloc(sizeof(LinestringList));
		/* populate the linestring list item */
		l->linestring = $1;
		l->next = NULL;
		/* set ring_list to the right thing */
		$$ = l;
	}
          | exterior_ring COMMA interior_rings
    {
		/* allocate a new linestring list item */
		LinestringList *l = (LinestringList *)malloc(sizeof(LinestringList));
		/* populate the linestring list item */
		l->linestring = $1;
		/* make the next point to the already constructed list */
		l->next = $3;
		/* set ring_list to the right thing */
	    $$ = l;
	}

polygon : POLYGON LPAREN ring_list RPAREN
    {
	    $$ = $3;
    }

pair_list: LPAREN pair RPAREN
    {
        /* allocate the new point list item */
		PointList *p = (PointList *)malloc(sizeof(PointList));
		/* populate the point list item */
		p->point = $2;
		p->next = NULL;
		/* set vertex_list to the right thing */
		$$ = p;
	}
         | pair_list COMMA LPAREN pair RPAREN 
    {
		/* retrieve the pointer to the beginning of the list */
		PointList *old = $1;
		/* declare a pointer to do traversal */
		PointList *oldptr;
		/* allocate a new point list item */
	    PointList *p = (PointList *)malloc(sizeof(PointList));

		/* populate the point list item */
		p->point = $4;
		p->next = NULL;
		/* traverse the current list until we reach the end */
		oldptr = old;
		while (oldptr->next != NULL) oldptr = oldptr->next;
		/* link in the next point list item */
		oldptr->next = p;
		/* set vertex_list to the right thing */
        $$ = old;
	}

multipoint : MULTIPOINT LPAREN pair_list RPAREN
    {
	    $$ = $3;
    }

linestring_list : LPAREN vertex_list RPAREN
    {
		/* allocate a new linestring list item */
		LinestringList *l = (LinestringList *)malloc(sizeof(LinestringList));
		/* populate the linestring list item */
		l->linestring = $2;
		l->next = NULL;
		/* set interior_rings to the right thing */
		$$ = l;
	}
                | linestring_list COMMA LPAREN vertex_list RPAREN
    {
		/* retrieve the pointer to the beginning of the list */
		LinestringList *old = $1;
		/* declare a pointer to do traversal */
		LinestringList *oldptr;
		/* allocate a new linestring list item */
		LinestringList *l = (LinestringList *)malloc(sizeof(LinestringList));

		/* populate the linestring list item */
		l->linestring = $4;
		l->next = NULL;
		/* traverse the current list until we reach the end */
		oldptr = old;
		while (oldptr->next != NULL) oldptr = oldptr->next;
		/* link in the next point list item */
		oldptr->next = l;
		/* set interior_rings to the right thing */
        $$ = old;
	}

multilinestring : MULTILINESTRING LPAREN linestring_list RPAREN
    {
	    $$ = $3;
    }

polygon_list : LPAREN ring_list RPAREN
    {
		/* allocate a new polygon list item */
	    PolygonList *p = (PolygonList *)malloc(sizeof(PolygonList));
		/* populate the polygon list item */
	    p->polygon = $2;
		p->next = NULL;
		/* set polygon_list to the right thing */
		$$ = p;
	}
             | polygon_list COMMA LPAREN ring_list RPAREN
    {
		/* retrieve the pointer to the beginning of the list */
		PolygonList *old = $1;
		/* declare a pointer to do traversal */
		PolygonList *oldptr;
		/* allocate a new polygon list item */
		PolygonList *p = (PolygonList *)malloc(sizeof(PolygonList));

		/* populate the polygon list item */
		p->polygon = $4;
		p->next = NULL;
		/* traverse the current list until we reach the end */
		oldptr = old;
		while (oldptr->next != NULL) oldptr = oldptr->next;
		/* link in the next point list item */
		oldptr->next = p;
		/* set polygon_list to the right thing */
        $$ = old;
	}

multipolygon : MULTIPOLYGON LPAREN polygon_list RPAREN
    {
	    $$ = $3;
    }

geometry_list : geometry
	{
        /* allocate the new geometry collection */
		GeometryCollection *g = 
		    (GeometryCollection *)malloc(sizeof(GeometryCollection));
		/* populate the new geometry collection */
		g->geometry = $1;
		g->next = NULL;
		/* set geometry list to the right thing */
		$$ = g;
	}
              | geometry_list COMMA geometry
	{
		/* stash the incoming list */
		GeometryCollection *old = $1;
		/* get a pointer for traversal */
		GeometryCollection *oldptr;
        /* allocate the new geometry collection */
		GeometryCollection *g = 
		    (GeometryCollection *)malloc(sizeof(GeometryCollection));

		/* populate the new geometry collection */
		g->geometry = $3;
		g->next = NULL;
		/* set the traversal pointer */
		oldptr = old;
		while (oldptr->next != NULL) oldptr = oldptr->next;
		/* link the new geometry collection item into the list */
		oldptr->next = g;
		/* set geometry list to the right thing */
		$$ = old;
	}

geometrycollection : GEOMETRYCOLLECTION LPAREN geometry_list RPAREN
    {
	    $$ = $3;
    }

%%
